-- DANOTATION -- interpreter and tabulator for danotation datasets
-- Copyright (c) 2011--2012 Alexandru Dan Corlan MD PhD
--
-- Webpage of this project: http://dan.corlan.net/software/danotation.html
-- short, persistent link: http://www.webcitation.org/64ewf3UTg
--
-- This program is free software, distributed under the GNU General Public Licence v2
--
-- History. Version 0.5 released 14 January 2012


with Ada.Characters;
with Ada.Characters.Handling; use Ada.Characters.Handling;
with Ada.Io_Exceptions;
with Ada.Text_Io; use Ada.Text_Io;
with Ada.Command_Line; use Ada.Command_Line;
with Ada.Calendar; use Ada.Calendar;

package Danlib is
   
   Toklen: constant Positive:= 1_000_000;
   
   subtype Intok is Positive range 1..Toklen;
   
   type Tokdata is array(Intok) of Character;
   
   type Danfile is record
      Fi: File_Type;
      Stdinp: Boolean:= True;
      Ix: Intok;
      Da: Tokdata;
      Eof: Boolean:= False;
      Timeval: Time;
      Stringp: Boolean:= False;
      Timep: Boolean:= False;
      Relativep: Boolean:= False;
      Crtc, Crtl: Long_Integer;
   end record;
   
   Strgslen: constant Positive:= 100_000;
   Nstrgsmax: constant Positive:= 1000;
   
   subtype Strgsbix is Integer range 1..Strgslen;
   subtype Strgsnix is Integer range 1..Nstrgsmax;
   type Strgsbuf is array(Integer range <>) of Character; 
   type Strgsix is array(Strgsnix) of Strgsbix;
   
   type Strings is record
      Bfill: Strgsbix:= 1;
      Nfill: Strgsnix:= 1;
      Buf: Strgsbuf(Strgsbix);
      Froms: Strgsix;
      Tos: Strgsix;
   end record;
   
   No_Label: constant Integer:= -999_999_999;
   
   type Danrecord is record
      Label: Integer:= No_Label;
      Names: Strings;
      Values: Strings;
   end record;
   
   procedure Reset(Strgs: in out Strings);
   function N(Strgs: in Strings) return Natural;
   procedure Addto(Strgs: in out Strings; Str: String);
   procedure Delelt(Strgs: in out Strings; Todel: Strgsnix);
   function Elt(Strgs: Strings; N: Positive) return String;
   
   procedure Open(F: in out Danfile; Nm: in String);   
   procedure Assign_To_Stdin(F: in out Danfile);
   procedure Close(F: in out Danfile);   
   procedure Read_Token(F: in out Danfile);
   procedure Reset(R: in out Danrecord);
   procedure Read(F: in out Danfile; R: in out Danrecord);
   function Token_Is_Label(F: Danfile) return Boolean;
   function Token_Is_Symbol(F: Danfile) return Boolean;
   function Token_String(F: Danfile) return String;
   function Locate(Strgs: Strings; S: String) return Natural;
   
   Eof_In_Sentence: exception;
   Syntax_Error: exception;
   No_Such_Attribute_Predefined: exception;
   Empty_Table_Head: exception;
   Wrong_Data: exception;
   No_Predicate_For_Table: exception;
   No_Subject_For_Table: exception;
   
   
  
end Danlib;

package body Danlib is
   
   procedure Open(F: in out Danfile; 
		  Nm: in String) is
      
   begin
      Open(F.Fi,In_File,Nm);
      F.Stdinp := False;
      F.Crtc := 0;
      F.Crtl := 1;
   end Open;
   
   procedure Assign_To_Stdin(F: in out Danfile) is
      
   begin
      F.Stdinp := True;
      F.Crtc := 0;
      F.Crtl := 1;
   end Assign_To_Stdin;
   
   procedure Close(F: in out Danfile) is
      
   begin
      Close(F.Fi);
   end Close;
   
   procedure Get(F: in out Danfile; C: in out Character) is
      
   begin
      if F.Stdinp then
	 Get_Immediate(Standard_Input, C);
      else
	 Get_Immediate(F.Fi, C);
      end if;
      F.Crtc := F.Crtc + 1;
      if C=Ascii.Lf then
	 F.Crtl := F.Crtl + 1;
      end if;
   end Get;
   
   pragma Inline(Get);
   
   function Is_Space(C: Character) return Boolean is
      
   begin
      return C=' ' or else C=Ascii.Lf;
   end Is_Space;
   
   procedure Read_Token(F: in out Danfile) is
      
      Ch: Character:= ' ';
      
   begin
      F.Eof := False;
      F.Ix := 1; -- next unwritten char
      F.Stringp := False;
      F.Timep :=  False;
      F.Relativep := False;
      loop
	  Get(F,Ch);
	  if not Is_Space(Ch) then
	     exit;
	  end if;
       end loop;
       F.Da(F.Ix) := Ch;
       if Ch='"' or Ch=''' then
	  F.Stringp := True;
       end if;
        F.Ix := F.Ix+1;
       loop
	  Get(F,Ch);
	 if Is_Space(Ch) and not F.Stringp then
	    exit;
	 end if;
	 if F.Stringp and then F.Da(1)=Ch then
	    -- maybe it ended
	    if F.Da(F.Ix-1)='\' then
	       F.Ix := F.Ix-1;
	    else
	       F.Da(F.Ix) := Ch;
	       return;
	    end if;
	 end if;
	 F.Da(F.Ix) := Ch;
	 F.Ix := F.Ix+1;
       end loop;
        F.Ix := F.Ix-1;
   exception
      when End_Error =>
	 if F.Ix=1 then
	    F.Eof := True;
	 else
	    F.Ix := F.Ix-1;
	 end if;
   end Read_Token;
   
   procedure Interpret_As_Time(F: in out Danfile) is -- totally rotten
      
      Ch: Character:= ' ';
      Timestyle: Character:= '0'; -- numeric only, standard fixed
      V: array(1..20) of Integer;
      C: array(1..20) of Character;
      Fl: Integer range 1..20:= 1;
      Yr,Mo,Dy: Natural:= 1;
      Hr,Mn,Sc,Ss: Natural:= 0;
      Fixends: String:= "//;::/";
      
   begin
      if Ch='@' then
	  F.Timep := True;
	  Get(F,Ch);
	  F.Relativep := Ch='+';
	  if F.Relativep then
	     Get(F,Ch);
	  end if;
	  if not Is_Digit(Ch) then 
	     Timestyle := Ch;
	     Get(F,Ch); -- we accept it, but do not parse for it yet
	  end if;
	     Fl := 1;
	  loop
	     -- timestyle '0'
	     -- read a vector of numbers sepparated by non-space non-digit characters
	     V(Fl):= 0;
	     while Is_Digit(Ch) loop
		V(Fl) := 10 * V(Fl) + Character'Pos(Ch) - Character'Pos('0');
		Get(F,Ch);
	     end loop;
	     C(Fl) := Ch;
	     if Is_Space(Ch) then
		exit;
	     end if;
	     Get(F,Ch);
	     Fl := Fl+1;
	  end loop;
	  Fl := Fl-1;
	  if F.Relativep then
	     null; -- not implemented yet
	  else
	     --must start with year
	     for I in 1..(Fl-1) loop
		if Fixends(I)/=C(I) then
		   raise Syntax_Error;
		end if;
	     end loop;
	     if Fl<1 then
		raise Syntax_Error; -- at least the year must be present
	     end if;
	     if Fl>=1 then
		Yr := V(1);
	     end if;
	     if Fl>=2 then
		Mo := V(2);
	     end if;
	     if Fl>=3 then
		Dy := V(3);
	     end if;
	     if Fl>=4 then
		Hr := V(4);
	     end if;
	     if Fl>=5 then
		Mn := V(5);
	     end if;
	     if Fl>=6 then
		Sc := V(6);
	     end if;
	     if Fl=7 then
		Ss := V(7);
	     end if;
--	     return Time_Of(Yr,Mo,Dy,Day_Duration(Hr*3600+Mn*60+Sc));
	  end if;
       end if;
   end Interpret_As_Time;
   
   function Token_Is_Label(F: Danfile) return Boolean is
     
   begin
      if F.Stringp or else F.Eof then
	 return False;
      end if;
      for I in 1..F.Ix loop
	 if not Is_Digit(F.Da(I)) then
	    return False;
	 end if;
      end loop;
      return True;
   end Token_Is_Label;
   
   function Token_Is_Symbol(F: Danfile) return Boolean is
     
   begin
      if F.Stringp or else F.Eof then
	 return False;
      end if;
      if not Is_Letter(F.Da(1)) and then F.Da(1)/='_' then
	 return False;
      end if;
      for I in 2..F.Ix loop
	 if not Is_Alphanumeric(F.Da(1)) and then F.Da(1)/='_' then
	    return False;
	 end if;
      end loop;
      return True;
   end Token_Is_Symbol;
   
   procedure Test_Read_Token(F: in out Danfile) is
      
   begin
      loop
	 Read_Token(F);
	 if F.Eof then
	    Put_Line("EOF");
	    exit;
	 end if;
	 if F.Stringp then
	    Put("String ");
	 end if;
	 if Token_Is_Label(F) then
	    Put("Label ");
	 end if;
	 if Token_Is_Symbol(F) then
	    Put("Symbol ");
	 end if;
	 
	 Put_Line(String(F.Da(1..F.Ix)) & "<");
      end loop;
   end Test_Read_Token;
   
   function Token_String(F: Danfile) return String is
      
   begin
      return String(F.Da(1..F.Ix));
   end Token_String;
   
   procedure Reset(R: in out Danrecord) is
      
   begin
      R.Label := No_Label;
      Reset(R.Names);
      Reset(R.Values);
   end Reset;
   
   procedure Read(F: in out Danfile; R: in out Danrecord) is
      
   begin
      Reset(R);
      Read_Token(F);
      if F.Eof then
	 return; -- with empty strings (N(R)=0)
      end if;
      if Token_Is_Label(F) then
	 R.Label := Integer'Value(Token_String(F));
	 Read_Token(F);
	 if F.Eof then
	    raise Eof_In_Sentence;
	 end if;
      end if;
      if Token_Is_Symbol(F) then
	 Addto(R.Names,"PREDICATE");
	 Addto(R.Values,Token_String(F));
	 Read_Token(F);
	 if F.Eof then
	    raise Eof_In_Sentence;
	 end if;
      else
	 raise Syntax_Error;
      end if;
      if Token_Is_Symbol(F) then
	 Addto(R.Names,"SUBJECT");
	 Addto(R.Values,Token_String(F));
	 Read_Token(F);
	 if F.Eof then
	    raise Eof_In_Sentence;
	 end if;
      else
	 raise Syntax_Error;
      end if;
      loop
	 if Token_Is_Symbol(F) then
	    if F.Ix=1 and then F.Da(1)='_' then
	       return;
	    end if;
	    Addto(R.Names,Token_String(F));
	    Read_Token(F);
	    Addto(R.Values,Token_String(F));
	    if F.Eof then
	       raise Eof_In_Sentence;
	    end if;
	    Read_Token(F);
	 else
	    raise Syntax_Error;
	 end if;
      end loop;
      
   end Read;      
   
   procedure Test_Read_Sentence(F: in out Danfile) is
      
   begin
      Read_Token(F);
      if F.Eof then
	 Put_Line("No more sentences");
	 return;
      end if;
      if Token_Is_Label(F) then
	 Put_Line("(label) " & Token_String(F));
	 Read_Token(F);
	 if F.Eof then
	    raise Eof_In_Sentence;
	 end if;
      end if;
      if Token_Is_Symbol(F) then
	 Put_Line("(predicate) " & Token_String(F));
	 Read_Token(F);
	 if F.Eof then
	    raise Eof_In_Sentence;
	 end if;
      else
	 raise Syntax_Error;
      end if;
      if Token_Is_Symbol(F) then
	 Put_Line("(subject) " & Token_String(F));
	 Read_Token(F);
	 if F.Eof then
	    raise Eof_In_Sentence;
	 end if;
      else
	 raise Syntax_Error;
      end if;
      loop
	 if Token_Is_Symbol(F) then
	    if F.Ix=1 and then F.Da(1)='_' then
	       Put_Line("End of sentence");
	       exit;
	    end if;
	    Put("  (attribute) " & Token_String(F));
	    Read_Token(F);
	    Put_Line(" = " & Token_String(F));
	    if F.Eof then
	       raise Eof_In_Sentence;
	    end if;
	    Read_Token(F);
	 else
	    raise Syntax_Error;
	 end if;
      end loop;
      
   end Test_Read_Sentence;
   
   procedure Check_Next_Sentence(F: in out Danfile) is
      
   begin
      Read_Token(F);
      if F.Eof then
	 return ;
      end if;
      if Token_Is_Label(F) then
	 Read_Token(F);
	 if F.Eof then
	    raise Eof_In_Sentence;
	 end if;
      end if;
      if Token_Is_Symbol(F) then
	 Read_Token(F);
	 if F.Eof then
	    raise Eof_In_Sentence;
	 end if;
      else
	 raise Syntax_Error;
      end if;
      if Token_Is_Symbol(F) then
	 Read_Token(F);
	 if F.Eof then
	    raise Eof_In_Sentence;
	 end if;
      else
	 raise Syntax_Error;
      end if;
      loop
	 if Token_Is_Symbol(F) then
	    if F.Ix=1 and then F.Da(1)='_' then
	       exit;
	    end if;
	    Read_Token(F);
	    if F.Eof then
	       raise Eof_In_Sentence;
	    end if;
	    Read_Token(F);
	 else
	    raise Syntax_Error;
	 end if;
      end loop;
      
   end Check_Next_Sentence;
   
   --  function Check_Danfile(F: in out Danfile) return Boolean is
      
   --  begin
   --     loop
   --  	 Check_Next_Sentence(F);
   --  	 if F.Eof then
   --  	    return True;
   --  	 end if;
   --     end loop;
   --  exception
   --     when others =>
   --  	 return False;
   --  end Check_Danfile;
   
   procedure Test_Read_File(F: in out Danfile) is
      
   begin
      loop
	 Test_Read_Sentence(F);
	 if F.Eof then
	    exit;
	 end if;
      end loop;
   end Test_Read_File;
   
   procedure Reset(Strgs: in out Strings) is
      
   begin
      Strgs.Nfill := 1;
      Strgs.Bfill := 1;
   end Reset;
   
   function N(Strgs: in Strings) return Natural is
      
   begin
      return Natural(Strgs.Nfill)-1;
   end N;
   
   procedure Addto(Strgs: in out Strings; Str: String) is
      
   begin
      Strgs.Buf(Strgs.Bfill..Strgs.Bfill+Str'Length-1) := Strgsbuf(Str);
      Strgs.Froms(Strgs.Nfill) := Strgs.Bfill;
      Strgs.Tos(Strgs.Nfill) := Strgs.Bfill+Str'Length-1;
      Strgs.Bfill := Strgs.Bfill + Str'Length;
      Strgs.Nfill := Strgs.Nfill + 1;
   end Addto;
   
   procedure Delelt(Strgs: in out Strings; Todel: Strgsnix) is
      
      L: Positive;
      
   begin
      L := Strgs.Tos(Todel)-Strgs.Froms(Todel)+1;
      if Todel < Strgs.Nfill-1 then
	 for I in Strgs.Froms(Todel+1)..Strgs.Bfill loop
	    Strgs.Buf(I-L) := Strgs.Buf(I);
	 end loop;
      end if;
      Strgs.Bfill := Strgs.Bfill - L;
      if Todel < Strgs.Nfill-1 then
	 for I in (Todel+1) .. Strgs.Nfill-1 loop
	    Strgs.Tos(I-1) := Strgs.Tos(I)-L;
	    Strgs.Froms(I-1) := Strgs.Froms(I)-L;
	 end loop;
      end if;
      Strgs.Nfill := Strgs.Nfill - 1;
   end Delelt;
   
   function Elt(Strgs: Strings; N: Positive) return String is
      
   begin
      return "" & String(Strgs.Buf(Strgs.Froms(N)..Strgs.Tos(N)));
   end Elt;
   
   function Locate(Strgs: Strings; S: String) return Natural is -- 0 if not found
      
   begin
      for I in 1..N(Strgs) loop
	 if S=Elt(Strgs,I) then
	    return I;
	 end if;
      end loop;
      return 0;
   end Locate;
   
   procedure Test_Strings is
      
      U: Strings;
      
   begin
      Reset(U);
      Addto(U,"one");
      Addto(U,"two");
      Addto(U,"three");
      Put_Line(Integer'Image(N(U)));
      for I in 1..3 loop
	 Put_Line(Elt(U,I));
      end loop;
   end Test_Strings;
   
   
end Danlib;


with Ada.Characters;
with Ada.Characters.Handling; use Ada.Characters.Handling;
with Ada.Io_Exceptions;
with Ada.Text_Io; use Ada.Text_Io;
with Ada.Command_Line; use Ada.Command_Line;
with Ada.Calendar; use Ada.Calendar;
with Danlib; use Danlib;

procedure Danotation is
   
   Inf: Danfile;
   
   -- the following are actually parameters to pi_sigma_extract
   
   Nar: Positive:= 1;
   Labels: Strings;
   Predicates: Strings;
   Subjects: Strings;
   Attributes: Strings;
   Values: Strings;
   Reports: Strings;
   Usage: exception;
   
   
   procedure Reset_Strings is
      
   begin
      Reset(Labels);
      Reset(Predicates);
      Reset(Subjects);
      Reset(Attributes);
      Reset(Values);
      Reset(Reports);
   end Reset_Strings;
   
   procedure Pi_Sigma_Extract is
      -- print a table of values of the attributes in the report
      -- for records that have the label, predicate and subject
      -- if specified (not empty string)
      -- do this by stream-processing the input
      
      Printit: Boolean:= True;
      Ls,Vs: Strings;
      
   begin
      for K in 1..N(Reports) loop -- table head
	 Put(Elt(Reports,K) & " ");
      end loop;
      if N(Reports)>0 then
	 New_Line;
      end if;
      
      loop -- for each input record, until return or exception
	 Printit := N(Reports)>0;
	 Read_Token(Inf);
	 Reset(Ls);
	 Reset(Vs);
	 if Inf.Eof then
	    return;
	 end if;
	 if Token_Is_Label(Inf) then
	    if N(Labels)>0 and then Elt(Labels,1)/=Token_String(Inf) then
	       Printit := False;
	    end if;
	    Addto(Ls,"LABEL");
	    Addto(Vs,Token_String(Inf));
	    Read_Token(Inf);
	    if Inf.Eof then
	       raise Eof_In_Sentence;
	    end if;
	 else
	    Addto(Ls,"LABEL");
	    Addto(Vs,"0");
	 end if;
	 if Token_Is_Symbol(Inf) then
	    if N(Predicates)>0 and then Elt(Predicates,1)/=Token_String(Inf) then
	       Printit := False;
	    end if;	    
	    Addto(Ls,"PREDICATE");
	    Addto(Vs,Token_String(Inf));
	    Read_Token(Inf);
	    if Inf.Eof then
	       raise Eof_In_Sentence;
	    end if;
	 else
	    raise Syntax_Error;
	 end if;
	 if Token_Is_Symbol(Inf) then
	    if N(Subjects)>0 and then Elt(Subjects,1)/=Token_String(Inf) then
	       Printit := False;
	    end if;	    
	    Addto(Ls,"SUBJECT");
	    Addto(Vs,Token_String(Inf));
	    Read_Token(Inf);
	    if Inf.Eof then
	       raise Eof_In_Sentence;
	    end if;
	 else
	    raise Syntax_Error;
	 end if;
	 
	 loop
	    if Token_Is_Symbol(Inf) then
	       if Inf.Ix=1 and then Inf.Da(1)='_' then
		  exit;
	       end if;
	       Addto(Ls,Token_String(Inf));
	       Read_Token(Inf);
	       Addto(Vs, Token_String(Inf));
	       if Inf.Eof then
		  raise Eof_In_Sentence;
	       end if;
	       for K in 1..N(Attributes) loop
		  if Elt(Ls,N(Ls))=Elt(Attributes,K) and then Elt(Vs,N(Vs))/=Elt(Values,K) then
		     Printit := False;
		  end if;
	       end loop;
	       Read_Token(Inf);
	    else
	       raise Syntax_Error;
	    end if;
	 end loop;
	 
	 if Printit then
	    for J in 1..N(Reports) loop
	       for K in 1..N(Ls) loop
		  if Elt(Reports,J)=Elt(Ls,K) then
		     Put(Elt(Vs,K) & " ");
		     exit;
		  end if;
	       end loop;
	    end loop;
	    New_Line;
	 end if;
      end loop;
      
   exception
      when Syntax_Error | Eof_In_Sentence =>
	 Put_Line(Standard_Error, "Syntax error at line: " & Long_Integer'Image(Inf.Crtl));
   end Pi_Sigma_Extract;
   
   
   procedure Preprocess(F: in out Danfile) is
      -- copy input to (standard) output, reformatted (pretty printed)
      -- perform processing of meta-statements such as 'include'
      -- or 'with', etc
      
      Cxnms,Cxvls: Strings;
      Nu,Nuc,Nut: Natural;
      Tabhead: Strings;
      Tabdata: Strings;
      
   begin
      Reset(Cxnms);
      Reset(Cxvls);
      Reset(Tabhead);
      loop -- for each input record, until return or exception
	 Read_Token(F);
	 if F.Eof then
	    return;
	 end if;
	 if Token_String(F)="end_data" then
	    return;
	 end if;
	 
	 if Token_String(F)="table_head" then
	    Reset(Tabhead);
	    loop
	       Read_Token(F);
	       if F.Eof then
		  raise Eof_In_Sentence;
	       end if;
	       if Token_String(F)="_" then
		  exit;
	       end if;
	       Addto(Tabhead, Token_String(F));
	    end loop;
	 elsif Token_String(F)="table_data" then
	    if N(Tabhead)<1 then
	       raise Empty_Table_Head;
	    end if;
	    Read_Token(F);
	    loop -- for each record
	       if Token_String(F)="end_table" then
		  exit;
	       end if;
	       Reset(Tabdata);
	       for I in 1..N(Tabhead) loop
		  if F.Eof then
		     raise Eof_In_Sentence;
		  end if;
		  if Token_String(F)="_" then
		     raise Wrong_Data;
		  end if;
		  Addto(Tabdata, Token_String(F));
	       Read_Token(F);
	       end loop;
	       if Token_String(F)/="_" then
		  raise Wrong_Data;
	       end if;
	       Read_Token(F);
	       
	       -- output the current record
	       Nuc := Locate(Cxnms,"PREDICATE");
	       if Nuc=0 then
		  Nut := Locate(Tabhead,"PREDICATE");
	       end if;
	       if Nuc=0 and then Nut=0 then
		  raise No_Predicate_For_Table;
	       end if;
	       if Nuc>0 then 
		  Put(Elt(Cxvls,Nuc) & " ");
	       else
		  Put(Elt(Tabdata,Nut) & " ");
	       end if;
	       
	       Nuc := Locate(Cxnms,"SUBJECT");
	       if Nuc=0 then
		  Nut := Locate(Tabhead,"SUBJECT");
	       end if;
	       if Nuc=0 and then Nut=0 then
		  raise No_Subject_For_Table;
	       end if;
	       if Nuc>0 then 
		  Put(Elt(Cxvls,Nuc) & " ");
	       else
		  Put(Elt(Tabdata,Nut) & " ");
	       end if;
	       
	       New_Line;
	       for I in 1..N(Cxnms) loop
		  if Elt(Cxnms,I)/="PREDICATE" and then Elt(Cxnms,I)/="SUBJECT" then
		     Put_Line("  " & Elt(Cxnms,I) & " " & Elt(Cxvls,I));
		  end if;
	       end loop;
	       
	       for I in 1..N(Tabhead) loop
		  if Elt(Tabhead,I)/="PREDICATE" and then Elt(Tabhead,I)/="SUBJECT" then
		     Put_Line("  " & Elt(Tabhead,I) & " " & Elt(Tabdata,I));
		  end if;
	       end loop;
	       Put_Line("_");
	       New_Line;
	    end loop;
	 elsif Token_String(F)="forget" then
	    Read_Token(F);
	    if F.Eof then
	       raise Eof_In_Sentence;
	    end if;
	    Nu := Locate(Cxnms,Token_String(F));
	    if Nu>0 then 
	       Delelt(Cxnms,Nu);
	       Delelt(Cxvls,Nu);
	    else
	       raise No_Such_Attribute_Predefined;
	    end if;
	 elsif Token_String(F)="with" then
	    Read_Token(F);
	    if F.Eof then
	       raise Eof_In_Sentence;
	    end if;
	    declare
	       Nm: String:= Token_String(F);
	    begin
	       Read_Token(F);
	       if F.Eof then
		  raise Eof_In_Sentence;
	       end if;
	       Nu := Locate(Cxnms,Nm);
	       if Nu>0 then
		  Delelt(Cxnms,Nu);
		  Delelt(Cxvls,Nu);
	       end if;
	       Addto(Cxnms,Nm);
	       Addto(Cxvls,Token_String(F));
	    end;
	 else
	    if Token_Is_Label(F) then
	       Put(Token_String(F) & " ");
	       Read_Token(F);
	       if F.Eof then
		  raise Eof_In_Sentence;
	       end if;
	    end if;
	    Nu := Locate(Cxnms, "PREDICATE");
	    if Nu>0 then -- must also check that the value is symbol, here
	       Put(Elt(Cxvls,Nu) & " ");
	    elsif Token_Is_Symbol(F) then
	       Put(Token_String(F) & " ");
	       Read_Token(F);
	       if F.Eof then
		  raise Eof_In_Sentence;
	       end if;
	    else
	       raise Syntax_Error;
	    end if;
	    Nu := Locate(Cxnms, "SUBJECT");
	    if Nu>0 then -- must also check that the value is symbol, here
	       Put(Elt(Cxvls,Nu) & " ");
	    elsif Token_Is_Symbol(F) then
	       Put(Token_String(F));
	       Read_Token(F);
	       if F.Eof then
		  raise Eof_In_Sentence;
	       end if;
	    else
	       raise Syntax_Error;
	    end if;
	    New_Line;
	    
	    for I in 1..N(Cxnms) loop
	       if Elt(Cxnms,I)/="PREDICATE" and then Elt(Cxnms,I)/="SUBJECT" then
		  Put("  " & Elt(Cxnms,I));
		  Put_Line(" " & Elt(Cxvls,I));
	       end if;
	    end loop;
	    
	    loop
	       if Token_Is_Symbol(F) then
		  if F.Ix=1 and then F.Da(1)='_' then
		     Put_Line("_");
		     New_Line;
		     exit;
		  end if;
		  Put("  " & Token_String(F)); 
		  Read_Token(F);
		  Put_Line(" " & Token_String(F));
		  if F.Eof then
		     raise Eof_In_Sentence;
		  end if;
		  Read_Token(F);
	       else
		  raise Syntax_Error;
	       end if;
	    end loop;
	 end if;
      end loop;
      
   exception
      when Syntax_Error | Eof_In_Sentence =>
	 Put_Line(Standard_Error, "Syntax error at line: " & Long_Integer'Image(F.Crtl));
   end Preprocess;
   
   Dr: Danrecord;
   
   procedure Dump is
      
      Nr: Natural:= 0;
      
   begin
      loop
	 Read(Inf, Dr);
	 if N(Dr.Names)=0 then
	    return;
	 end if;
	 Nr := Nr+1;
	 for I in 1..N(Dr.Names) loop
	    Put_Line(Integer'Image(Nr) &
		       Integer'Image(I) &
		       " " & Elt(Dr.Names,I) &
		       " " & Elt(Dr.Values,I));
	 end loop;
      end loop;
   end Dump;
   
   -- the following are actually parameters to pi_sigma_extract
   
--   Nar: Positive:= 1;
--   Labels: Strings;
--   Predicates: Strings;
--   Subjects: Strings;
--   Attributes: Strings;
--   Values: Strings;
--   Reports: Strings;
--   Usage: exception;
   
   
   type Pisigma_Output_Mode is (Dump, Table); -- more in the future
   
   procedure Multi_Pi_Sigma(Mode: Pisigma_Output_Mode;
			    Header: Boolean:= False;
			    Repsep: String:= ",";
			    Colsep: String:=" ";
			    Nonaval: String:= "NA") is
      
      Nr,Nin: Natural:= 0;
      
      function Record_Fits return Boolean is
	 P,S,Crf: Boolean:= False;
      begin
	 if N(Predicates)=0 then
	    P := True;
	 end if;
	 for I in 1..N(Predicates) loop
	    if Elt(Predicates,I)=Elt(Dr.Values,1) then
	       P := True;
	    end if;
	 end loop;
	 if not P then
	    return False;
	 end if;
	 
	 if N(Subjects)=0 then
	    S := True;
	 end if;
	 for I in 1..N(Subjects) loop
	    if Elt(Subjects,I)=Elt(Dr.Values,2) then
	       S := True;
	    end if;
	 end loop;
	 if not S then
	    return False;
	 end if;
	 
	 for I in 1..N(Attributes) loop
	    Crf := False;
	    for K in 1..N(Dr.Names) loop
	       if Elt(Attributes,I)=Elt(Dr.Names,K) then
		  if Elt(Values,I)=Elt(Dr.Values,K) then
		     Crf := True;
		  end if;
	       end if;
	    end loop;
	    if Crf = False then
	       return False;
	    end if;
	 end loop;
	 
	 return True;
	 
      end Record_Fits;
      
   begin
      if Mode=Table and then Header then
	 for K in 1..N(Reports) loop
	    if K>1 then
	       Put(Colsep);
	    end if;
	    Put(Elt(Reports,K));
	 end loop;
	 New_Line;
      end if;
      loop
	 Read(Inf, Dr);
	 if N(Dr.Names)=0 then
	    return;
	 end if;
	 Nr := Nr+1;
	 if Record_Fits then
	    case Mode is
	       when Dump =>
		  for I in 1..N(Dr.Names) loop
		     Put_Line(Integer'Image(Nr) &
				Integer'Image(I) &
				" " & Elt(Dr.Names,I) &
				" " & Elt(Dr.Values,I));
		  end loop;
	       when Table =>
		  for J in 1..N(Reports) loop
		     Nin := 0;
		     if J>1 then
			Put(Colsep);
		     end if;
		     for K in 1..N(Dr.Names) loop
			if Elt(Reports,J)=Elt(Dr.Names,K) then
			   if Nin>0 then
			      Put(Repsep);
			   end if;
			   Put(Elt(Dr.Values,K));
			   Nin := Nin+1;
			end if;
		     end loop;
		     if Nin=0 then
			Put(Nonaval);
		     end if;
		  end loop;
		  New_Line;
	    end case;
	 end if;
      end loop;
   end Multi_Pi_Sigma;
   
   Dumpmode: Pisigma_Output_Mode:= Table;
   Add_Header: Boolean:= False;

begin
   Assign_To_Stdin(Inf);
   Reset_Strings;
   Nar := 1;
   while Nar<=Argument_Count loop
      if Argument(Nar)="-i" or else Argument(Nar)="--interpret" then
	 Preprocess(Inf);
	 return;
      elsif Argument(Nar)="-d" or else Argument(Nar)="--dump" then
	 Dumpmode := Dump;
	 Nar := Nar+1;
      elsif Argument(Nar)="-h" or else Argument(Nar)="--header" then
	 Add_Header := True;
	 Nar := Nar+1;
      elsif Argument(Nar)="-p" or else Argument(Nar)="--predicate" then -- predicate
	 if Nar=Argument_Count then
	    raise Usage;
	 end if;
	 Addto(Predicates,Argument(Nar+1));
	 Nar := Nar+2;
      elsif Argument(Nar)="-s" or else Argument(Nar)="--subject" then -- subject
	 if Nar=Argument_Count then
	    raise Usage;
	 end if;
	 Addto(Subjects,Argument(Nar+1));
	 Nar := Nar+2;
      elsif Argument(Nar)="-l" or else Argument(Nar)="--label" then -- label
	 if Nar=Argument_Count then
	    raise Usage;
	 end if;
	 Addto(Labels,Argument(Nar+1));
	 Nar := Nar+2;
      elsif Argument(Nar)="-eq" or else Argument(Nar)="--equals" then -- value
	 if Nar+1>=Argument_Count then
	    raise Usage;
	 end if;
	 Addto(Attributes,Argument(Nar+1));
	 Addto(Values,Argument(Nar+2));
	 Nar := Nar+3;
      else
	 Addto(Reports,Argument(Nar));
	 Nar := Nar+1;
      end if;
   end loop;
   Multi_Pi_Sigma(Dumpmode,Header => Add_Header);
end Danotation;
